/*
 * Routine to write out a row of the path matrix using a minimum amount of
 * disk space.
 * 
 * This is achieved by creating two files, one containing the nonzero elements
 * of the row and the other containing integers identifying the number of
 * nonzero elements and the the indexes of the nonzero elements
 * 
 * subscripting begins with 0 so if you are using vectors that begin at one use
 * &vector[1] as the argument.
 * 
 * ifile contains the integers, ffile contains the floats
 * 
 * 
 * PutrowP - put all positive values PutRow0 - put all nonzero values
 * fabs(value) > effective_zero
 * 
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include "invsubs.h"

#define ABSOLUTE(x) ( (x > 0) ? x : -x)

void PutrowP(FILE *intfile, FILE *flofile, float *theRow, int nc, float effective_zero)
{
	int             c, nwritten;
	int             n_notZero, *indices;
	int             nBytes;
	float          *buffer, temp;

	indices = (int *) malloc((unsigned) (nc * sizeof(int)));
	buffer = (float *) malloc((unsigned) (nc * sizeof(int)));

	n_notZero = 0;
	for (c = 0; c < nc; c++) {
		temp = theRow[c];
		if (temp > effective_zero) {
			indices[n_notZero] = c;
			buffer[n_notZero] = temp;
			n_notZero += 1;
		}
	}

	nBytes = sizeof(int);
	nwritten = write(fileno(intfile), &n_notZero, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the number of indices\n");
		exit(-1);
	}
	nBytes *= n_notZero;
	nwritten = write(fileno(intfile), indices, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the indices\n");
		exit(-1);
	}
	nBytes = n_notZero * sizeof(float);
	nwritten = write(fileno(flofile), buffer, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the floats.\n");
		exit(-1);
	}
	free((char *) buffer);
	free((char *) indices);

	return;
}

void Putrow0(FILE *intfile, FILE *flofile, float *theRow, int nc, float effective_zero)
{
	int             c, nwritten;
	int             n_notZero, *indices;
	int             nBytes;
	float          *buffer, temp;

	indices = (int *) malloc((unsigned) (nc * sizeof(int)));
	buffer = (float *) malloc((unsigned) (nc * sizeof(int)));

	n_notZero = 0;
	for (c = 0; c < nc; c++) {
		temp = theRow[c];
		if (ABSOLUTE(temp) > effective_zero) {
			indices[n_notZero] = c;
			buffer[n_notZero] = temp;
			n_notZero += 1;
		}
	}

	nBytes = sizeof(int);
	nwritten = write(fileno(intfile), &n_notZero, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the number of indices\n");
		exit(-1);
	}
	nBytes *= n_notZero;
	nwritten = write(fileno(intfile), indices, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the indices\n");
		exit(-1);
	}
	nBytes = n_notZero * sizeof(float);
	nwritten = write(fileno(flofile), buffer, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the floats.\n");
		exit(-1);
	}
	free((char *) buffer);
	free((char *) indices);

	return;
}

/*
 * Write out selected values of a row - the floating point values are stored
 * in "theFloats" and the corresponding indices are in "theIndices".  The
 * number of values to put is contained int n_notZero.  This routine assumes
 * you have already decided what values to save on the disk.  This function
 * begins at the address of the arrays so if you are using a subscripting of
 * one, give &buffer[1],&indices[1] as the arguments in the calling function.
 * Also, the indexes will begin at zero for the array so you must account for
 * this in the calling function.
 * 
 */
void PutRowValues(FILE *intfile, FILE *flofile, float *theFloats, int *theIndices, int *n_notZero)
{
	int             c, nwritten;
	int             nBytes;
	float          *buffer, temp;

	nBytes = sizeof(int);
	nwritten = write(fileno(intfile), n_notZero, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the number of indices\n");
		exit(-1);
	}
	nBytes *= *n_notZero;
	nwritten = write(fileno(intfile), theIndices, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the indices\n");
		exit(-1);
	}
	nBytes = *n_notZero * sizeof(float);
	nwritten = write(fileno(flofile), theFloats, nBytes);
	if (nwritten != nBytes) {
		fprintf(stderr, "Message from Putrow:\n");
		fprintf(stderr, "Problem writing the floats.\n");
		exit(-1);
	}
	return;
}

/*
 * Routine to read a row of the path matrix stored using a minimum amount of
 * disk space.
 * 
 * This is achieved by creating two files, one containing the nonzero elements
 * of the row and the other containing integers identifying the number of
 * nonzero elements and the the indexes of the nonzero elements
 * 
 * subscripting begins with 0 so if you are using vectors that begin at one use
 * &vector[1] as the argument.
 * 
 * ifile contains the integers, ffile contains the floats
 * 
 */

void Getrow(FILE *intfile, FILE *flofile, float *theRow, int nc, int *ierror)
{
	int             c, nread;
	int             nBytes;
	int             n_notZero, *indices;
	float          *buffer;

	*ierror = 0;
	for (c = 0; c < nc; c++)
		theRow[c] = 0.0;

	nBytes = sizeof(int);

	nread = read(fileno(intfile), &n_notZero, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the number of non-zero indices.\n");
		*ierror = -1;
		return;
	}
	if (n_notZero == 0) {
		return;
	}
	indices = (int *) malloc((unsigned) (n_notZero * sizeof(int)));
	buffer = (float *) malloc((unsigned) (n_notZero * sizeof(float)));

	nBytes *= n_notZero;
	nread = read(fileno(intfile), indices, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the indices.\n");
		*ierror = -1;
		return;
	}
	nBytes = n_notZero * sizeof(float);
	nread = read(fileno(flofile), buffer, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the floating point values.\n");
		*ierror = -1;
		return;
	}
	for (c = 0; c < n_notZero; c++)
		theRow[indices[c]] = buffer[c];

	free((char *) buffer);
	free((char *) indices);

	return;
}
/*
 * Routine to read a row of the path matrix stored using a minimum amount of
 * disk space.
 * 
 * This version does not unpack the row but passes the array of nonzero indices
 * (theIndices) and the floating point values (theFloats) and the number of
 * nozero values (nnzero) to the calling function
 * 
 * This is achieved by creating two files, one containing the nonzero elements
 * of the row and the other containing integers identifying the number of
 * nonzero elements and the the indexes of the nonzero elements
 * 
 * subscripting begins with 0 so if you are using vectors that begin at one use
 * &vector[1] as the argument.
 * 
 * ifile contains the integers, ffile contains the floats
 * 
 */

void GetRowValues(FILE *intfile, FILE *flofile, float *theFloats, int *theIndices, int *nnzero, int *ierror)
{
	int             c, nread;
	int             nBytes;
	int             n_notZero, *indices;
	float          *buffer;

	nBytes = sizeof(int);

	*ierror = 0;

	nread = read(fileno(intfile), &n_notZero, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the number of non-zero indices.\n");
		*ierror = -1;
		return;
	}
	if (n_notZero == 0) {
		*nnzero = 0;
		return;
	}
	nBytes *= n_notZero;
	nread = read(fileno(intfile), theIndices, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the indices.\n");
		*ierror = -1;
		return;
	}
	nBytes = n_notZero * sizeof(float);
	nread = read(fileno(flofile), theFloats, nBytes);
	if (nread != nBytes) {
		fprintf(stderr, "Message from Getrow:\n");
		fprintf(stderr, "Problem reading in the floating point values.\n");
		*ierror = -1;
		return;
	}
	*nnzero = n_notZero;

	return;
}
