/****************************************************************************************/
/*
 * Program: mnmlsqr Version: 1.0
 * 
 * program to invert group velocities using great-circle paths This code uses a
 * conjugate gradient algorithm (LSQR) to perform an approximate inversion of
 * the matrix equations.
 * 
 * Chuck Ammon St. Louis University, October 1996 (nmlsqr)
 *    reworked under Sun's workshop November 2002 (nmlsqr)
 *    converted to multiple-periods    March 2003 (mnmlsqr)
 * 
 * Some Details:
 * 
 * The inversion grid is set up to be uniform length cells (measured in km)
 * assuming a sinusoidal projection). The output is ascii percent velocity
 * perturbations, the fast and the slow values are written into separate
 * files. The procedure for the inversion is iterative, but the ray paths are
 * fixed during the inversion.
 * 
 * The input are in ascii format and consist of the event and station locations
 * and the phase velocity.  The initial model is a constant velocity model,
 * which you must also supply.
 * 
 * Laplacian smoothness constraints are included and so is an aprior model
 * constraint (minimize the difference between the solution and a specified
 * model), which right now is limited to using the constant velocity model.
 * 
 * 
 */
/****************************************************************************************/

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include "invsubs.h"
#include "libget.h"
#include <unistd.h>

#define deg_to_rad 0.0174532
#define kSmall 0.0001

/* prototypes */
/*
void setpar(int ac, char **av) ;
void mstpar(char *name,char *type,int *val);
int getpar(char *name, char *type, int *val);
*/

/****************************************************************************/
/* define a structure to hold the observations */
/****************************************************************************/


void getStats(float *x, int n, float *mean, float *variance);

/****************************************************************************/

int main(ac, av)
	int             ac;
	char          **av;
{
	float           az;
	float          *l, *l0;
	float          *v, *s, *ds, *dc_velocity, *s0;
	float           time_tol = 0.0001;
	float           small = .00005;
	float           damper = 0.1;
	float           minres = 2.0;
	float           sum = 0, res, predicted;
	float          *plats, *plons, *pYs, *pXs, smple_dist, max_cond_num = 100;
	float           recs_constraint = 1;
	float           Wx, Wy;
	float           minlat, minlon, observed, rms;
	float           *initialPeriodMisfits, *finalPeriodMisfits, *sumWtsPerPeriod,temp;
	float           center_lon, eastern, western, northern, southern, min_x;
	float           lsqr_res, corr_wt = 1;
	float           maxdist = 0, mindist = 0;
	float           min_res_bin, res_bin_width;
	float           sum_of_wts = 0, epsilon = 1.0;
	float           centerlat, so_far, fraction_done;
	float           arc_distance, distance, *observations, smoother = 1, apmodelwt = 0;
	float           vel_res, vel_sum, vel_rms,initialRMS,initialVelocityRMS;
	float          *theResiduals,theInitialMean,theInitialVariance,theFinalMean,theFinalVariance;
	float           *ltt, *ltl, *modelLength, *modelRoughness, rough;
	float           minPeriod, dPeriod, distcheck, suggestedWeight, wres;

	char            obsfile[256], resfile[256], pathMatrix[256], iModel[256];
	FILE           *rcfd;
	FILE           *intfile, *flofile;
	FILE           *fd0, *fd1, *fd2, *fd3, *logFile;

	int             nCells, nCellsPerImage, nImages, verbose = 0, write_times = 0;
	int             nx, ny, i, j, k, n, fd;
	int             numRows, nobs, iperiod, ioffset, *nObsPerPeriod;
	int             lsqr_iter = 20, iobs;
	int             iteration, max_iteration = 1;
	int             nsegs, mx_segs = 20000;
	int            *indices, nnot_zero;
	int             computePaths = 1, r_error;
	int             nwritten = 0;
	int            *resbins, nres_bins, ires_bin;
	int             iwt = 0;
	int             keep = 1, ekeep = 1, mask = 0, minobs = 4;

	struct OBSERVATION anObservation;

	logFile = fopen("SRFLSQR.log", "w");
	setlinebuf(logFile);
	strcpy(pathMatrix, "DUMMY");

	/**********************************************************************************/
	/* this value is used to sample along the path - it is in km */
	/* the max path length is smpl_dist * mx_segs, which is 20,000km right now */
	/**********************************************************************************/
	smple_dist = 1.0;
	mx_segs = 20000;

	/*****************************************************/
	/* set up a vector to hold the binned residual stats */
	/*****************************************************/
	res_bin_width = 2;
	nres_bins = 501;
	min_res_bin = -501;
	resbins = (int *) calloc(nres_bins, sizeof(int));

	/*****************************************************/
	/* User input: */
	/**************************************************** */
	setpar(ac, av);
	getpar("max_iteration", "d", &max_iteration);
	/*mstpar("dc_velocity", "f", &dc_velocity);*/
	mstpar("xwidth", "f", &Wx);
	mstpar("ywidth", "f", &Wy);
	mstpar("minlat", "f", &minlat);
	mstpar("minlon", "f", &minlon);
	mstpar("nx", "d", &nx);
	mstpar("ny", "d", &ny);
	mstpar("nobs", "d", &nobs);
	mstpar("obsfile", "s", obsfile);
/*
	mstpar("imodel","s",iModel);
*/
	getpar("verbose", "d", &verbose);
	getpar("predict", "d", &write_times);
	getpar("damper", "f", &damper);
	getpar("computepaths", "d", &computePaths);
	computePaths = 1; /* the optimal dc_velocity is calculated while computing paths */
	getpar("epsilon", "f", &epsilon);
	getpar("iwtevents", "d", &iwt);
	getpar("max_cond_num", "f", &max_cond_num);
	getpar("lsqr_iter", "d", &lsqr_iter);
	getpar("smoother", "f", &smoother);
	getpar("apmodelwt", "f", &apmodelwt);

/*
	mstpar("minPeriod","f", &minPeriod);
	mstpar("dperiod",  "f", &dPeriod);
*/
	mstpar("nImages", "d", &nImages); /* i.e. nPeriods (or later, R & G) */
	

				     
	/*********************************************************************************/
	/* Echo the User input */
	/*********************************************************************************/
	fprintf(logFile, "xwidth=%f ywidth=%f\n", Wx, Wy);
	fprintf(logFile, "minlat=%f minlon=%f\n", minlat, minlon);
	fprintf(logFile, "nx=%d ny=%d nobs=%d\n", nx, ny, nobs);
	fprintf(logFile, "nobs=%d\n", nobs);
	fprintf(logFile, "obsfile=%s\n", obsfile);
	fprintf(logFile, "damper=%f\n", damper);
	if (iwt == 1)
		fprintf(stdout, "Weighting the events using event bins.\n");

	/*********************************************************************************/
	/* Allocate memory for the other vectors and arrays  */
	/*********************************************************************************/
	nCellsPerImage = nx * ny;
	nCells = nCellsPerImage * nImages;

	observations = (float *) calloc(nobs + 2 * nCells, sizeof(float));
	theResiduals = (float *) calloc(nobs, sizeof(float));

	l0      = (float *) calloc(nCellsPerImage, sizeof(float));
	l       = (float *) calloc(nCells, sizeof(float));
	indices = (int *)   calloc(nCells, sizeof(int));
	ds      = (float *) calloc(nCells, sizeof(float));
	s       = (float *) calloc(nCells + 1, sizeof(float));
	v       = (float *) calloc(nCells, sizeof(float));
	plats   = (float *) calloc(mx_segs, sizeof(float));
	plons   = (float *) calloc(mx_segs, sizeof(float));
	pXs     = (float *) calloc(mx_segs, sizeof(float));
	pYs     = (float *) calloc(mx_segs, sizeof(float));

	s0      = (float *) calloc(nCells + 1, sizeof(float));

	ltt =         (float *) calloc(nImages, sizeof(float));
	ltl =         (float *) calloc(nImages, sizeof(float));
	dc_velocity = (float *) calloc(nImages, sizeof(float));
	modelLength = (float *) calloc(nImages, sizeof(float));
	modelRoughness = (float *) calloc(nImages, sizeof(float));
	initialPeriodMisfits = (float *) calloc(nImages, sizeof(float));
	finalPeriodMisfits = (float *) calloc(nImages, sizeof(float));
	sumWtsPerPeriod = (float *) calloc(nImages, sizeof(float));
	nObsPerPeriod = (int *) calloc(nImages, sizeof(float));

	/*********************************************************************************/
	/* read in the initial model (velocity) */
	/*********************************************************************************/
/*
	fd2 = fopen(iModel,"r");
	fread(s0,sizeof(float),nCellsPerImage,fd2);
	fclose(fd2);
*/
/* instead of reading in the velocity model, we will introduce spikes
   on an  8x8 grid 
   We will also use two velocities  3.0 +- 10%  = 2.70, 3.30 */
	for(i=0; i < ny ; i++){
		for(j=0; j < nx ; j++){
			k = j + i*nx;
			if( i%8 == 0 && j%8  == 0){
                                if( ((i+j)/8)%2 == 0 ){
					s0[k] = 2.70;
				} else {
					s0[k] = 3.30;
				}
			} else {
				s0[k] = 3.00;
			}
			
		}
	}

	/****************************************************************************/
	/* set up some constants for the grid. */
	/****************************************************************************/

	northern = minlat + (ny * Wy) / 111.19;
	southern = minlat;
	western = minlon;
	eastern = minlon + (nx * Wx) / 111.19;
	center_lon = 0.5 * (eastern + western);
	min_x = (center_lon - western) * 111.19;
	centerlat = 0.5 * (southern + northern);
	fprintf(logFile, "Grid:\nsouthern,northern,western,eastern\n");
	fprintf(logFile, "%f,%f,%f,%f\n", southern, northern, western, eastern);

	iteration = 1;

	if (computePaths) {
		/****************************************************************************/
		/* open the pathmatrix files for output of ALL Ray paths */
		/****************************************************************************/

		intfile = fopen("L_indices", "w");
		if (intfile == NULL) {
			fprintf(stderr, "Problem opening the path index file.\n");
			exit(-1);
		}
		flofile = fopen("L_floats", "w");
		if (flofile == NULL) {
			fprintf(stderr, "Problem opening the path float file.\n");
			exit(-1);
		}
		nwritten = 0;

		/****************************************************************************/
		/* open up the observation and predictions files */
		/****************************************************************************/

		fd0 = fopen(obsfile, "r");
		if (fd0 == NULL) {
			fprintf(stderr, "Problem opening file: %s\n", obsfile);
			exit(-1);
		}
		/****************************************************************************/
		/* loop over the observations */
		/****************************************************************************/

		if (verbose) {
			fprintf(stdout, "------------------------------\n");
			fprintf(stdout, "Computing the ray paths\n");
			fprintf(stdout, "------------------------------\n");
		}
		fraction_done = 0.05;
		fprintf(stderr, "xxxxxxxxxxxxxxxxxxxx\n");

		for (iobs = 0; iobs < nobs; iobs++) {

			anObservation = nReadAnObservation(fd0);

			delta_az(&anObservation.evlat, &anObservation.evlon, &anObservation.stlat, &anObservation.stlon, &arc_distance, &az);
			arc_distance *= 111.19;
			nsegs = (arc_distance / smple_dist) + 1;

			/* used to compute the least-squares slowness */
			ltl[anObservation.iperiod] += arc_distance * arc_distance ;
			ltt[anObservation.iperiod] += arc_distance * arc_distance / anObservation.phs_velocity;

			if (nsegs > mx_segs)
				nsegs = mx_segs;

			Gcarc(anObservation.evlat, anObservation.evlon, anObservation.stlat, anObservation.stlon, az, arc_distance, plats, plons, nsegs);

			for (i = 0; i < nsegs; i++) {
				LATLON_to_XY(plats[i], plons[i], &pXs[i], &pYs[i], minlat, center_lon, min_x);
/*
printf("nsegs %d %d %f %f %f %f\n",nsegs,i,plats[i], plons[i],pXs[i],pYs[i]);
*/
			}

			GetGCARCLengths(pXs, pYs, plats, plons, nsegs, nx, ny, Wx, Wy, l0);

			ioffset = anObservation.iperiod * nCellsPerImage;

			for(i=0;i<nCells;i++)
			  l[i] = 0;

			for(i = 0; i< nCellsPerImage;i++)
			  l[i + ioffset] = (l0[i]);


			PutrowP(intfile, flofile, l, nCells, (float) 0.1);
			nwritten += 1;

			/* user feedback on how long to go */
			so_far = (float) iobs / (float) nobs;
			if (so_far > fraction_done) {
				fprintf(stderr, "x");
				fraction_done += 0.05;
			}
		}		/* end of loop over observations */
		fprintf(stdout, "o\n");
		fclose(intfile);
		fclose(flofile);
		fclose(fd0);
		fprintf(stdout, "Number of rows written: %d\n", nwritten);
	}

	/* finished with the path computations */


	/* compute the least squares slowness value and store its inverse in dc_velocity  */
	fd2 = fopen("mean_velocity.txt","w");
	fprintf(fd2,"____________________________\n");
	fprintf(fd2,"\nInitial DC Velocities\n");
	fprintf(fd2,"____________________________\n");
	for(j=0;j<nImages;j++)
	  {
	    /****************************************************************************/
	    /* Initialize the velocity model */
	    /****************************************************************************/
	    for (i = 0; i < nCellsPerImage; i++) {
	           k = i + j * nCellsPerImage;
	        s[k] = 1.0 / s0[i];
	        v[k] = s0[i];
	       ds[k] = 0;
	    }
	  }
	fprintf(fd2,"____________________________\n");
	fclose(fd2);

	/********************************************************************************/
	/*
	 * NOW COMPUTE THE NEW TRAVEL TIMES
	 */
	/********************************************************************************/

	intfile = fopen("L_indices", "r");
	if (intfile == NULL) {
		fprintf(stderr, "Problem opening the path index file.\n");
		fprintf(stderr, "Check the computepaths input parameter.\n");
		fprintf(stderr, "I'm quitting.\n");
		exit(-1);
	}
	flofile = fopen("L_floats", "r");
	if (flofile == NULL) {
		fprintf(stderr, "Problem opening the path float file.\n");
		exit(-1);
	}
	/********************************************************************************/
	/* open the observation file */
	/********************************************************************************/

	fd0 = fopen(obsfile, "r");
	if (fd0 == NULL) {
		fprintf(stderr, "Problem opening file: %s\n", obsfile);
		exit(-1);
	}
	numRows = 0;
	sum = 0;
	vel_sum = 0;

	fd2 = fopen("predictedValues.txt","w");
	fd3 = fopen("new.disp","w");
	/***************************************************************************************************************************/
	/* set up the vector with the times or observations storage */
	/***************************************************************************************************************************/

	for (iobs = 0; iobs < nobs; iobs++) {

		anObservation = nReadAnObservation(fd0);
		delta_az(&anObservation.evlat, &anObservation.evlon, &anObservation.stlat, &anObservation.stlon, &arc_distance, &az);
		arc_distance *= 111.19;
		observed = arc_distance / anObservation.phs_velocity;


		GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		if (r_error == -1)
			fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);

		predicted = 0;
		distance = 0;
		for (j = 0; j < nnot_zero; j++) {
			predicted += l[j] * s[indices[j]];
			ds[indices[j]] += l[j];	/* store the "coverage" in ds (still includes weighting factor)*/
			distance += l[j];
		}


		distcheck = fabs(distance - arc_distance)/arc_distance;

		fprintf(fd2,"%-32s %10.3f %10.3f %10.3f %10.3f %10.3f %6.2f %10.3f\n",
			anObservation.event_id,
			anObservation.evlat, anObservation.evlon, anObservation.stlat,
			anObservation.stlon, arc_distance / predicted,anObservation.period,anObservation.weight);	      
		fprintf(fd3,"%s %10.4f %10.4f %10.4f %10.4f %10.4f %6.2f %10.3f %d\n",
                        anObservation.event_id,
                        anObservation.evlat, anObservation.evlon, anObservation.stlat,
                        anObservation.stlon, arc_distance/predicted,anObservation.period,anObservation.weight,anObservation.iperiod);


		numRows += 1;

	}			/* end of loop over observations */

	fclose(fd0);
	fclose(fd2);
	fclose(fd3);
	fclose(intfile);
	fclose(flofile);

	/*************************************************************************************************/
	/* write the model into slow and fast ascii files (only need to create one file) */
	/*************************************************************************************************/

	nImages = 1; 
	WriteModels(iteration,ds,s,v,minlat,minlon,Wx,Wy,nx,ny,center_lon,min_x,nImages);

	fclose(logFile);

	return (0);
}
/* end of main */

